#include <fstream.h>
#include <stdlib.h>
#include <alloc.h>
#include <conio.h>
#include <math.h>
#include "vector.h"

#define R           64
#define R_2         128
#define R_BITS      6
#define R2          128
#define R3          96
#define EDGE        32
#define HEIGHT      20
#define MAX_X       319
#define MAX_Y       199
#define TEX_SIZE    64
#define TEX_BITS    6
#define TEX_MASK    63
#define NORMALS     9
#define OUTER_COLOR 0x00

char far *texture;
char far *tex_normals;
char far *light_map[NORMALS];
int offs_x = 0, offs_y = 0;
Vector normals[NORMALS] = {Vector(1, 0, 1), Vector(0, 1, 1), Vector(-1, 0, 1),
                           Vector(0, -1, 1), Vector(1, 1, 1), Vector(-1, 1, 1),
                           Vector(-1, -1, 1), Vector(1, -1, 1), Vector(0, 0, 1)};

const double C = M_PI / 180;

class Light {
  public:
    Vector org;
    Light() { org = 0; }
    Light(Vector org) { Light::org = org; }
} light(Vector(MAX_X / 2, MAX_Y / 2, HEIGHT));

void PutPixel(int x, int y, char c)
{
  asm
  {
      push   0A000h
      pop    es
      mov    di,  y
      mov    ax,  di
      shl    di,  8
      shl    ax,  6
      add    di,  ax
      add    di,  x
      mov    al,  c
      stosb
  }
}

void SetMode(int mode)
{
  asm
  {
      mov    ax,  mode
      int    10h
  }
}

void SetPal(char c, char r, char g, char b)
{
  asm
  {
      mov    dx,  03C8h
      mov    al,  c
      out    dx,  al
      inc    dx
      mov    al,  r
      out    dx,  al
      mov    al,  g
      out    dx,  al
      mov    al,  b
      out    dx,  al
  }
}

void BuildLightMap()
{
  double first_light, spot_light, r;

  for (int n = 0; n < NORMALS; n++)
  for (int i = -R; i < R; i++)
    for (int j = -R; j < R; j++)
    {
      first_light = Normalize(normals[n]) & Normalize(Vector(0, 0, light.org.z) - Vector(i, j, 0));
      if (first_light > 1) first_light = 1;
      if (first_light < 0) first_light = 0;
      r = sqrt(i * i + j * j);
      if (r < R - EDGE)
        spot_light = first_light;
      else
        if (r > R)
          spot_light = 0;
        else
          spot_light = first_light * (R - r) / EDGE;
      light_map[n][(j + R) * R_2 + i + R] = spot_light * 62 + random(2);
    }
}

void LoadTexture(char *filename)
{
  int index;

  ifstream in(filename, ios::binary);
  in.seekg(1078, ios::beg);
  in.read((char *)texture, TEX_SIZE * TEX_SIZE);
  in.close();

  for (int i = 0; i < TEX_SIZE; i++)
    for (int j = 0; j < TEX_SIZE; j++)
    {
      index = (j << TEX_BITS) + i;
      if (texture[index] == OUTER_COLOR)
        texture[index] = 1;
      else
        if (i & (TEX_SIZE >> 1) ^ j & (TEX_SIZE >> 1))
          texture[index] = 0;
        else
          texture[index] = 2;
    }
}

void SetTexel(int i, int j, int normal)
{
  int index = (j << TEX_BITS) + i;

  if (i > TEX_SIZE - 1) i -= TEX_SIZE;
  if (i < 0) i += TEX_SIZE;
  if (j > TEX_SIZE - 1) j -= TEX_SIZE;
  if (j < 0) j += TEX_SIZE;

  if (texture[index] != 1)
    tex_normals[index] = normal;
}

void BuildTextureNormals()
{
  int index;

  for (int i = 0; i < TEX_SIZE; i++)
    for (int j = 0; j < TEX_SIZE; j++)
      tex_normals[(j << TEX_BITS) + i] = 8;

  for (int i = 0; i < TEX_SIZE; i++)
  {
    tex_normals[i] = 3;
    tex_normals[(((TEX_SIZE >> 1) - 1) << TEX_BITS) + i] = 1;
    tex_normals[((TEX_SIZE >> 1) << TEX_BITS) + i] = 3;
    tex_normals[((TEX_SIZE - 1) << TEX_BITS) + i] = 1;
    tex_normals[(i << TEX_BITS)] = 2;
    tex_normals[(i << TEX_BITS) + (TEX_SIZE >> 1) - 1] = 0;
    tex_normals[(i << TEX_BITS) + (TEX_SIZE >> 1)] = 2;
    tex_normals[(i << TEX_BITS) + TEX_SIZE - 1] = 0;
  }

  tex_normals[0] = 6;
  tex_normals[((TEX_SIZE >> 1) - 1) << TEX_BITS] = 5;
  tex_normals[(TEX_SIZE >> 1) << TEX_BITS] = 6;
  tex_normals[(TEX_SIZE - 1) << TEX_BITS] = 5;

  tex_normals[(TEX_SIZE >> 1) - 1] = 7;
  tex_normals[(((TEX_SIZE >> 1) - 1) << TEX_BITS) + (TEX_SIZE >> 1) - 1] = 4;
  tex_normals[((TEX_SIZE >> 1) << TEX_BITS) + (TEX_SIZE >> 1) - 1] = 7;
  tex_normals[((TEX_SIZE - 1) << TEX_BITS) + (TEX_SIZE >> 1) - 1] = 4;

  tex_normals[TEX_SIZE >> 1] = 6;
  tex_normals[(((TEX_SIZE >> 1) - 1) << TEX_BITS) + (TEX_SIZE >> 1)] = 5;
  tex_normals[((TEX_SIZE >> 1) << TEX_BITS) + (TEX_SIZE >> 1)] = 6;
  tex_normals[((TEX_SIZE - 1) << TEX_BITS) + (TEX_SIZE >> 1)] = 5;

  tex_normals[TEX_SIZE - 1] = 7;
  tex_normals[(((TEX_SIZE >> 1) - 1) << TEX_BITS) + TEX_SIZE - 1] = 4;
  tex_normals[((TEX_SIZE >> 1) << TEX_BITS) + TEX_SIZE - 1] = 7;
  tex_normals[((TEX_SIZE - 1) << TEX_BITS) + TEX_SIZE - 1] = 4;

  for (int i = 0; i < TEX_SIZE; i++)
    for (int j = 0; j < TEX_SIZE; j++)
    {
      index = (j << TEX_BITS) + i;
      if (texture[index] == 1)
      {
        SetTexel(i + 1, j, 0 + 16);
        SetTexel(i, j + 1, 1 + 16);
        SetTexel(i - 1, j, 2 + 16);
        SetTexel(i, j - 1, 3 + 16);
        SetTexel(i + 1, j + 1, 4 + 16);
        SetTexel(i - 1, j + 1, 5 + 16);
        SetTexel(i - 1, j - 1, 6 + 16);
        SetTexel(i + 1, j - 1, 7 + 16);
        tex_normals[index] = 8;
      }
    }

  for (int i = 0; i < TEX_SIZE; i++)
    for (int j = 0; j < TEX_SIZE; j++)
    {
      index = (j << TEX_BITS) + i;
      if (tex_normals[index] >= 16)
      {
        texture[index] = 1;
        tex_normals[index] -= 16;
      }
    }
}

void AsmRender(int x1, int x2, int y1, int y2, int a, int b, int offs_x, int offs_y)
{
  asm
  {
      push   0A000h
      pop    es
      mov    di,  y1
      mov    ax,  di
      shl    di,  8
      shl    ax,  6
      add    di,  ax
      add    di,  x1
      mov    cx,  y2
      sub    cx,  y1
@@1:  push   cx
      mov    dx,  y2
      sub    dx,  cx
      mov    cx,  x2
      sub    cx,  x1
@@2:  mov    bx,  dx
      add    bx,  offs_y
      and    bx,  TEX_MASK
      shl    bx,  TEX_BITS
      mov    ax,  x2
      sub    ax,  cx
      add    ax,  offs_x
      and    ax,  TEX_MASK
      add    bx,  ax
      push   ds
      lds    si,  texture
      add    si,  bx
      lodsb
      mov    ah,  al
      lds    si,  tex_normals
      add    si,  bx
      lodsb
      pop    ds
      push   ax
      mov    bx,  dx
      sub    bx,  b
      and    bl,  0FEh
      shl    bx,  R_BITS
      mov    ax,  x2
      sub    ax,  cx
      sub    ax,  a
      shr    ax,  1
      add    bx,  ax
      pop    ax
      push   cx
      mov    cx,  bx
      xor    bx,  bx
      mov    bl,  al
      shl    bl,  2
      push   ds
      lds    si,  light_map[bx]
      add    si,  cx
      lodsb
      pop    ds
      pop    cx
      shl    ah,  6
      xchg   bl,  ah
      add    bl,  al
      xchg   bl,  al
      stosb
      loop   @@2
      pop    cx
      sub    di,  x2
      add    di,  x1
      add    di,  320
      loop   @@1
  }
}

void Render()
{
  int x1 = light.org.x - R_2, a = x1;
  int x2 = light.org.x + R_2;
  int y1 = light.org.y - R_2, b = y1;
  int y2 = light.org.y + R_2;
  int light;

  if (x1 < 0) x1 = 0;
  if (x2 > MAX_X) x2 = MAX_X;
  if (y1 < 0) y1 = 0;
  if (y2 > MAX_Y) y2 = MAX_Y;

  AsmRender(x1, x2, y1, y2, a, b, offs_x, offs_y);
}

void main()
{
  double r, a;

  randomize();
  SetMode(0x13);

  for (int i = 0; i < NORMALS; i++)
    light_map[i] = (char far *)farmalloc(R * R * 4);

  texture = new char far[TEX_SIZE * TEX_SIZE];
  tex_normals = new char far[TEX_SIZE * TEX_SIZE];

  BuildLightMap();
  LoadTexture("lambda.bmp");
  BuildTextureNormals();

  for (int i = 0; i < 64; i++)
  {
    SetPal(i, i, i, i);
    SetPal(i + 64, i, i, 0);
    SetPal(i + 128, 0, 0, i);
  }

  do
  {
    Render();
    r = sin(a * M_PI / 90);
    light.org.x = cos(a * C) * r * R2 + MAX_X / 2;
    light.org.y = sin(a * C) * r * R3 + MAX_Y / 2;
    a++;
    if (a > 360)
      a -= 360;
    offs_x++;
    offs_y++;
    offs_x &= TEX_MASK;
    offs_y &= TEX_MASK;
  } while (!kbhit());

  delete texture;
  delete tex_normals;

  for (int i = 0; i < NORMALS; i++)
    farfree(light_map[i]);

  SetMode(0x03);
}
